@@ -1,6 +1,6 @@
 /* pkl-ast.h - Abstract Syntax Tree for Poke.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -61,7 +61,8 @@ enum pkl_ast_code
   PKL_AST_FORMAT_ARG,
   PKL_AST_INCRDECR,
   PKL_AST_GCD,
-  PKL_AST_LAST_EXP = PKL_AST_GCD,
+  PKL_AST_ASM_EXP,
+  PKL_AST_LAST_EXP = PKL_AST_ASM_EXP,
   /* Types.  */
   PKL_AST_TYPE,
   PKL_AST_STRUCT_TYPE_FIELD,
@@ -83,13 +84,14 @@ enum pkl_ast_code
   PKL_AST_LOOP_STMT_ITERATOR,
   PKL_AST_RETURN_STMT,
   PKL_AST_EXP_STMT,
-  PKL_AST_TRY_CATCH_STMT,
-  PKL_AST_TRY_UNTIL_STMT,
+  PKL_AST_TRY_STMT,
+  PKL_AST_TRY_STMT_BODY,
+  PKL_AST_TRY_STMT_HANDLER,
   PKL_AST_PRINT_STMT,
-  PKL_AST_BREAK_STMT,
-  PKL_AST_CONTINUE_STMT,
+  PKL_AST_BREAK_CONTINUE_STMT,
   PKL_AST_RAISE_STMT,
-  PKL_AST_LAST_STMT = PKL_AST_RAISE_STMT,
+  PKL_AST_ASM_STMT,
+  PKL_AST_LAST_STMT = PKL_AST_ASM_STMT,
   PKL_AST_LAST
 };
 
@@ -363,7 +365,7 @@ pkl_ast_node pkl_ast_make_string (pkl_ast ast,
 
 /* PKL_AST_ARRAY nodes represent array literals.  Each array holds a
    sequence of elements, all of them having the same type.  There must
-   be at least one element in the array, i.e. emtpy arrays are not
+   be at least one element in the array, i.e. empty arrays are not
    allowed.  */
 
 #define PKL_AST_ARRAY_NELEM(AST) ((AST)->array.nelem)
@@ -452,8 +454,8 @@ pkl_ast_node pkl_ast_make_struct_field (pkl_ast ast,
                                         pkl_ast_node name,
                                         pkl_ast_node exp);
 
-/* PKL_AST_EXP nodes represent unary and binary expressions,
-   consisting on an operator and one or two operators, respectively.
+/* PKL_AST_EXP nodes represent expressions consisting on an operator
+   and one or more operators.
 
    The supported operators are specified in pkl-ops.def.
    The supported attributes are defined in pkl-attrs.def.
@@ -479,7 +481,7 @@ struct pkl_ast_exp
   enum pkl_ast_op code;
   enum pkl_ast_attr attr;
   uint8_t numops : 8;
-  union pkl_ast_node *operands[2];
+  union pkl_ast_node *operands[3];
   int flag;
 };
 
@@ -492,8 +494,17 @@ pkl_ast_node pkl_ast_make_binary_exp (pkl_ast ast,
                                       pkl_ast_node op1,
                                       pkl_ast_node op2);
 
+pkl_ast_node pkl_ast_make_ternary_exp (pkl_ast ast,
+                                       enum pkl_ast_op code,
+                                       pkl_ast_node op1,
+                                       pkl_ast_node op2,
+                                       pkl_ast_node op3);
+
 const char *pkl_attr_name (enum pkl_ast_attr attr);
 
+pkl_ast_node pkl_ast_handle_bconc_ass_stmt (pkl_ast ast,
+                                            pkl_ast_node exp);
+
 /* PKL_AST_COND_EXP nodes represent conditional expressions, having
    exactly the same semantics than the C tertiary operator:
 
@@ -589,12 +600,6 @@ pkl_ast_node pkl_ast_make_enum (pkl_ast ast,
    BODY is a PKL_AST_COMP_STMT node containing the statements that
    conform the function body.
 
-   NFRAMES is a counter used by the parser as an aid in determining
-   the number of lexical frames a RETURN_STMT should pop before
-   returning from the function.  While parsing, this contains the
-   number of frames pushed to the environment at any moment.  After
-   parsing, this field is not used anymore.
-
    NARGS is the number of formal arguments.
 
    NAME is a C string containing the name used to declare the
@@ -607,7 +612,6 @@ pkl_ast_node pkl_ast_make_enum (pkl_ast ast,
 #define PKL_AST_FUNC_ARGS(AST) ((AST)->func.args)
 #define PKL_AST_FUNC_FIRST_OPT_ARG(AST) ((AST)->func.first_opt_arg)
 #define PKL_AST_FUNC_BODY(AST) ((AST)->func.body)
-#define PKL_AST_FUNC_NFRAMES(AST) ((AST)->func.nframes)
 #define PKL_AST_FUNC_NAME(AST) ((AST)->func.name)
 #define PKL_AST_FUNC_NARGS(AST) ((AST)->func.nargs)
 #define PKL_AST_FUNC_METHOD_P(AST) ((AST)->func.method_p)
@@ -623,7 +627,6 @@ struct pkl_ast_func
   union pkl_ast_node *body;
 
   int nargs;
-  int nframes;
   char *name;
   int method_p;
   pvm_program program;
@@ -715,8 +718,10 @@ pkl_ast_node pkl_ast_make_trimmer (pkl_ast ast,
 
    BASE must point to a PKL_AST_ARRAY node.
 
-   INDEX must point to an expression whose evaluation is the offset of
-   the element into the field, in units of the field's SIZE.
+   INDEX must point to an expression whose evaluation is either an
+   integer with the 0 base index of the indexed element, or an offset
+   indicating the exact offset from the beginning of the array to the
+   indexed element.
 
    IS_INDEXED is a boolean indicating whether the indexer is
    immediately indexed with another [] operator.  This is used in the
@@ -778,8 +783,22 @@ pkl_ast_node pkl_ast_make_struct_ref (pkl_ast ast,
    ENDIAN is the endianness to use when reading and writing data
    to/from the field.
 
-   OPTCOND is a boolean expression that, if present, specifies whether
-   the field exists in the struct or not.
+   OPTCOND_PRE is a boolean expression that, if present, specifies
+   whether the field exists in the struct or not.  The expression is
+   evaluated before the field has been mapped or constructed and thus
+   it cannot reference the field itself.
+
+   OPTCOND_POST is a boolean expression that, if present, specifies
+   whether the field exists in the struct or not.  The expression is
+   evaluated after the field has been mapped or constructed and thus
+   it can reference the field itself.
+
+   COMPUTED_P is a flag that is set when the field is a computed
+   field, i.e. it is a field whose values are computed using a pair of
+   struct methods.  Computed fields are not mapped nor constructed the
+   usual way.  If COMPUTED_P is true then INITIALIZER, CONSTRAINT,
+   LABEL, OPTCOND_PRE and OPTCOND_POST should be NULL, and also ENDIAN
+   should be PKL_AST_ENDIAN_DFL.
 
    INITIALIZER is an expression, that will be used to derive an
    implicit constraint, and also as the initialization value for this
@@ -790,9 +809,15 @@ pkl_ast_node pkl_ast_make_struct_ref (pkl_ast ast,
 #define PKL_AST_STRUCT_TYPE_FIELD_TYPE(AST) ((AST)->sct_type_elem.type)
 #define PKL_AST_STRUCT_TYPE_FIELD_SIZE(AST) ((AST)->sct_type_elem.size)
 #define PKL_AST_STRUCT_TYPE_FIELD_CONSTRAINT(AST) ((AST)->sct_type_elem.constraint)
+#define PKL_AST_STRUCT_TYPE_FIELD_CONSTRAINT_SRC(AST) ((AST)->sct_type_elem.constraint_src)
 #define PKL_AST_STRUCT_TYPE_FIELD_LABEL(AST) ((AST)->sct_type_elem.label)
 #define PKL_AST_STRUCT_TYPE_FIELD_ENDIAN(AST) ((AST)->sct_type_elem.endian)
-#define PKL_AST_STRUCT_TYPE_FIELD_OPTCOND(AST) ((AST)->sct_type_elem.optcond)
+#define PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE(AST) ((AST)->sct_type_elem.optcond_pre)
+#define PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST(AST) ((AST)->sct_type_elem.optcond_post)
+#define PKL_AST_STRUCT_TYPE_FIELD_OPTIONAL_P(AST) \
+  (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE (AST)    \
+   || PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST (AST))
+#define PKL_AST_STRUCT_TYPE_FIELD_COMPUTED_P(AST) ((AST)->sct_type_elem.computed_p)
 #define PKL_AST_STRUCT_TYPE_FIELD_INITIALIZER(AST) ((AST)->sct_type_elem.initializer)
 
 struct pkl_ast_struct_type_field
@@ -805,8 +830,11 @@ struct pkl_ast_struct_type_field
   union pkl_ast_node *constraint;
   union pkl_ast_node *initializer;
   union pkl_ast_node *label;
-  union pkl_ast_node *optcond;
+  union pkl_ast_node *optcond_pre;
+  union pkl_ast_node *optcond_post;
   int endian;
+  int computed_p;
+  char *constraint_src;
 };
 
 pkl_ast_node pkl_ast_make_struct_type_field (pkl_ast ast,
@@ -816,7 +844,8 @@ pkl_ast_node pkl_ast_make_struct_type_field (pkl_ast ast,
                                              pkl_ast_node initializer,
                                              pkl_ast_node label,
                                              int endian,
-                                             pkl_ast_node optcond);
+                                             pkl_ast_node optcond_pre,
+                                             pkl_ast_node optcond_post);
 
 /* PKL_AST_FUNC_TYPE_ARG nodes represent the arguments part of a
    function type.
@@ -875,14 +904,15 @@ pkl_ast_node pkl_ast_make_func_type_arg (pkl_ast ast,
    is to avoid processing the same type more than once.
 
    In struct types, NELEM is the number of elements in the struct
-   type, NFIELD is the number of fields, and NDECL is the number of
-   declarations.  ELEMS is a chain of elements, which can be
+   type, NFIELD is the number of non-computed fields, NCFIELD is the
+   number of computed fields, and NDECL is the number of declarations.
+   ELEMS is a chain of elements, which can be
    PKL_AST_STRUCT_TYPE_FIELD or PKL_AST_DECL nodes, potentially mixed.
    PINNED_P is 1 if the struct is pinned, 0 otherwise.  MAPPER, WRITER
    CONSTRUCTOR, FORMATER, PRINTER, COMPARATOR, INTEGRATOR and
-   DEINTEGRATOR are used to hold closures, or PVM_NULL.  ITYPE, if
-   not NULL, is an AST node with an integral type, that defines the
-   nature of this struct type as integral.
+   DEINTEGRATOR are used to hold closures, or PVM_NULL.  ITYPE, if not
+   NULL, is an AST node with an integral type, that defines the nature
+   of this struct type as integral.
 
    In offset types, BASE_TYPE is a PKL_AST_TYPE with the base type for
    the offset's magnitude, and UNIT is either a PKL_AST_IDENTIFIER
@@ -915,10 +945,9 @@ pkl_ast_node pkl_ast_make_func_type_arg (pkl_ast ast,
 #define PKL_AST_TYPE_A_WRITER(AST) ((AST)->type.val.array.closures[1])
 #define PKL_AST_TYPE_A_BOUNDER(AST) ((AST)->type.val.array.closures[2])
 #define PKL_AST_TYPE_A_CONSTRUCTOR(AST) ((AST)->type.val.array.closures[3])
-#define PKL_AST_TYPE_A_PRINTER(AST) ((AST)->type.val.array.closures[4])
-#define PKL_AST_TYPE_A_FORMATER(AST) ((AST)->type.val.array.closures[5])
-#define PKL_AST_TYPE_A_INTEGRATOR(AST) ((AST)->type.val.array.closures[6])
+#define PKL_AST_TYPE_A_INTEGRATOR(AST) ((AST)->type.val.array.closures[4])
 #define PKL_AST_TYPE_S_NFIELD(AST) ((AST)->type.val.sct.nfield)
+#define PKL_AST_TYPE_S_NCFIELD(AST) ((AST)->type.val.sct.ncfield)
 #define PKL_AST_TYPE_S_NDECL(AST) ((AST)->type.val.sct.ndecl)
 #define PKL_AST_TYPE_S_NELEM(AST) ((AST)->type.val.sct.nelem)
 #define PKL_AST_TYPE_S_ELEMS(AST) ((AST)->type.val.sct.elems)
@@ -928,12 +957,10 @@ pkl_ast_node pkl_ast_make_func_type_arg (pkl_ast ast,
 #define PKL_AST_TYPE_S_MAPPER(AST) ((AST)->type.val.sct.closures[0])
 #define PKL_AST_TYPE_S_WRITER(AST) ((AST)->type.val.sct.closures[1])
 #define PKL_AST_TYPE_S_CONSTRUCTOR(AST) ((AST)->type.val.sct.closures[2])
-#define PKL_AST_TYPE_S_PRINTER(AST) ((AST)->type.val.sct.closures[3])
-#define PKL_AST_TYPE_S_COMPARATOR(AST) ((AST)->type.val.sct.closures[4])
-#define PKL_AST_TYPE_S_INTEGRATOR(AST) ((AST)->type.val.sct.closures[5])
-#define PKL_AST_TYPE_S_FORMATER(AST) ((AST)->type.val.sct.closures[6])
-#define PKL_AST_TYPE_S_DEINTEGRATOR(AST) ((AST)->type.val.sct.closures[7])
-#define PKL_AST_TYPE_S_TYPIFIER(AST) ((AST)->type.val.sct.closures[8])
+#define PKL_AST_TYPE_S_COMPARATOR(AST) ((AST)->type.val.sct.closures[3])
+#define PKL_AST_TYPE_S_INTEGRATOR(AST) ((AST)->type.val.sct.closures[4])
+#define PKL_AST_TYPE_S_DEINTEGRATOR(AST) ((AST)->type.val.sct.closures[5])
+#define PKL_AST_TYPE_S_TYPIFIER(AST) ((AST)->type.val.sct.closures[6])
 #define PKL_AST_TYPE_S_ITYPE(AST) ((AST)->type.val.sct.itype)
 #define PKL_AST_TYPE_O_UNIT(AST) ((AST)->type.val.off.unit)
 #define PKL_AST_TYPE_O_BASE_TYPE(AST) ((AST)->type.val.off.base_type)
@@ -977,6 +1004,7 @@ struct pkl_ast_type
     {
       size_t nelem;
       size_t nfield;
+      size_t ncfield;
       size_t ndecl;
       union pkl_ast_node *elems;
       union pkl_ast_node *itype;
@@ -1059,6 +1087,20 @@ pkl_ast_node pkl_type_integral_promote (pkl_ast ast,
                                         pkl_ast_node type1,
                                         pkl_ast_node type2);
 
+/* Return the PKL_AST_STRUCT_TYPE_FIELD with the given FIELD_NAME in
+   the given STRUCT_TYPE.  If no such a field exists in the struct
+   type, return NULL.  */
+
+pkl_ast_node pkl_ast_get_struct_type_field (pkl_ast_node struct_type,
+                                            const char *field_name);
+
+/* Return the PKL_AST_DECL node node corresponding to the method named
+   METHOD_NAME in the given STRUCT_TYPE.  If no such method exists
+   then return NULL.  */
+
+pkl_ast_node pkl_ast_get_struct_type_method (pkl_ast_node struct_type,
+                                             const char *method_name);
+
 /* Return an expression that evaluates to an increment step for the
    given TYPE.  If the provided type doesn't support the notion of
    increment step this function returns NULL.  */
@@ -1075,6 +1117,10 @@ pkl_ast_node pkl_ast_type_incr_step (pkl_ast ast, pkl_ast_node type);
    NAME is PKL_AST_IDENTIFIER node containing the name in the
    association.
 
+   PREV_NAME is used in order to cache the name in the declaration
+   when the later is set to "" in re-definition of non-immutable
+   global objects.  See pkl-env.c to see how this field is used.
+
    INITIAL is the initial value of the entity.  The kind of node
    depends on what is being declared:
    - An expression node for a variable.
@@ -1091,16 +1137,20 @@ pkl_ast_node pkl_ast_type_incr_step (pkl_ast ast, pkl_ast_node type);
    whatever.
 
    STRUCT_FIELD_P indicates whether this declaration is for a variable
-   corresponding to a struct field.  */
+   corresponding to a struct field.
+
+   IMMUTABLE_P indicates whether this declaration can be redefined.
+   Used when bootstrapping the compiler.  */
 
 #define PKL_AST_DECL_KIND(AST) ((AST)->decl.kind)
 #define PKL_AST_DECL_NAME(AST) ((AST)->decl.name)
-#define PKL_AST_DECL_TYPE(AST) ((AST)->decl.type)
+#define PKL_AST_DECL_PREV_NAME(AST) ((AST)->decl.prev_name)
 #define PKL_AST_DECL_INITIAL(AST) ((AST)->decl.initial)
 #define PKL_AST_DECL_ORDER(AST) ((AST)->decl.order)
 #define PKL_AST_DECL_SOURCE(AST) ((AST)->decl.source)
 #define PKL_AST_DECL_STRUCT_FIELD_P(AST) ((AST)->decl.struct_field_p)
 #define PKL_AST_DECL_IN_STRUCT_P(AST) ((AST)->decl.in_struct_p)
+#define PKL_AST_DECL_IMMUTABLE_P(AST) ((AST)->decl.immutable_p)
 
 #define PKL_AST_DECL_KIND_ANY 0
 #define PKL_AST_DECL_KIND_VAR 1
@@ -1115,7 +1165,9 @@ struct pkl_ast_decl
   int kind;
   int struct_field_p;
   int in_struct_p;
+  int immutable_p;
   char *source;
+  char *prev_name;
   union pkl_ast_node *name;
   union pkl_ast_node *initial;
   int order;
@@ -1362,7 +1414,7 @@ pkl_ast_node pkl_ast_make_format (pkl_ast ast, pkl_ast_node fmt,
 
    EXP is an expression node evaluating to the value to format.
 
-   BASE is the numeration base to use when formating this argument.
+   BASE is the numeration base to use when formatting this argument.
 
    BEGIN_SC, if not NULL, marks that this argument is a %<class>
    directive, and is a NULL-terminated string with the name of the
@@ -1376,7 +1428,31 @@ pkl_ast_node pkl_ast_make_format (pkl_ast ast, pkl_ast_node fmt,
    value of EXP, respectively.
 
    VALUE_P indicates whether the argument shall be printed as a PVM
-   value or not (whether this arguement corresponds to a %v or not).
+   value or not (whether this argument corresponds to a %v or not).
+
+   FLOATING_POINT_P indicates whether the argument shall be interpreted
+   as a floating-point number.  The argument should be an integral
+   value.  The argument will be casted to UINT<32> for 32-bit
+   floating-point numbers and will be casted to UINT<64> for 64-bit
+   floating-point numbers.
+
+   FLOATING_POINT_STYLE specify the floating-point style; the valid
+   values are 'f', 'e' and 'g' ASCII characters.  'f' corresponds to
+   %f-tag which print the number in the style [-]ddd.ddd where
+   the number of digits after the decimal-point character is equal to
+   the precision.  'e' corresponds to %e-tag which print the number in
+   the style [-]d.ddde+-dd (so called scientific notation).  And finally
+   'g' which corresponds to %g-tag will choose either %f or %e based on
+   length of the representation.
+
+   FLOATING_POINT_PREC specify the precision in string representation
+   of the floating-point number.  For %f- and %g-tags it specify the
+   number of digits after the decimal-point character, and for %g-tag
+   it specify the number of significant digits.
+
+   FLOATING_POINT_WIDTH specify the width of the floating-point number
+   which can be one the PKL_AST_FLOATING_POINT_WIDTH_* constants defined
+   below.
 
    FORMAT_MODE and FORMAT_DEPTH specify how the argument shall be
    printed if VALUE_P is true.  FORMAT_MODE can be one of the
@@ -1389,12 +1465,23 @@ pkl_ast_node pkl_ast_make_format (pkl_ast ast, pkl_ast_node fmt,
 #define PKL_AST_FORMAT_ARG_BEGIN_SC(AST) ((AST)->format_arg.begin_sc)
 #define PKL_AST_FORMAT_ARG_END_SC(AST) ((AST)->format_arg.end_sc)
 #define PKL_AST_FORMAT_ARG_VALUE_P(AST) ((AST)->format_arg.value_p)
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_P(AST)     \
+  ((AST)->format_arg.floating_point_p)
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_STYLE(AST) \
+  ((AST)->format_arg.floating_point_style)
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_PREC(AST)  \
+  ((AST)->format_arg.floating_point_prec)
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH(AST) \
+  ((AST)->format_arg.floating_point_width)
 #define PKL_AST_FORMAT_ARG_FORMAT_MODE(AST) ((AST)->format_arg.format_mode)
 #define PKL_AST_FORMAT_ARG_FORMAT_DEPTH(AST) ((AST)->format_arg.format_depth)
 
 #define PKL_AST_FORMAT_MODE_FLAT 0
 #define PKL_AST_FORMAT_MODE_TREE 1
 
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH_SINGLE 0
+#define PKL_AST_FORMAT_ARG_FLOATING_POINT_WIDTH_DOUBLE 1
+
 struct pkl_ast_format_arg
 {
   struct pkl_ast_common common;
@@ -1403,6 +1490,10 @@ struct pkl_ast_format_arg
   char *end_sc;
   int base;
   int value_p;
+  int floating_point_p;
+  int floating_point_style;
+  unsigned floating_point_prec;
+  int floating_point_width;
   int format_mode;
   int format_depth;
   char *suffix;
@@ -1490,6 +1581,33 @@ struct pkl_ast_lambda
 
 pkl_ast_node pkl_ast_make_lambda (pkl_ast ast, pkl_ast_node function);
 
+/* PKL_AST_ASM nodes represent asm expressions.
+
+   TEMPLATE is a string node contaning assembler instructions.
+
+   TYPE is the type of the element that is left at the top of
+   the stack once the program in TEMPLATE is executed.
+
+   INPUTS is a chain of expressions.  The result of these expressions
+   will be pushed in the stack in the given order before executing
+   TEMPLATE.  */
+
+#define PKL_AST_ASM_EXP_TEMPLATE(AST) ((AST)->asm_exp.template)
+#define PKL_AST_ASM_EXP_TYPE(AST) ((AST)->asm_exp.type)
+#define PKL_AST_ASM_EXP_INPUTS(AST) ((AST)->asm_exp.inputs)
+
+struct pkl_ast_asm_exp
+{
+  struct pkl_ast_common common;
+
+  union pkl_ast_node *template;
+  union pkl_ast_node *type;
+  union pkl_ast_node *inputs;
+};
+
+pkl_ast_node pkl_ast_make_asm_exp (pkl_ast ast, pkl_ast_node type,
+                                   pkl_ast_node template, pkl_ast_node inputs);
+
 /* PKL_AST_INCRDECR nodes represent {pre,post}{increment,decrement}
    expressions.
 
@@ -1532,53 +1650,21 @@ pkl_ast_node pkl_ast_make_incrdecr (pkl_ast ast,
 
    If BUILTIN is not PKL_AST_BUILTIN_NONE, then this compound
    statement is a compiler builtin, i.e. specific code will be
-   generated for this node.  In this case, STMTS should be NULL.  */
+   generated for this node.  In this case, STMTS should be NULL.
+
+   If FRAMELESS_P is true then the compound statement will not use its
+   own lexical frame.  */
 
 #define PKL_AST_COMP_STMT_STMTS(AST) ((AST)->comp_stmt.stmts)
 #define PKL_AST_COMP_STMT_BUILTIN(AST) ((AST)->comp_stmt.builtin)
 #define PKL_AST_COMP_STMT_NUMVARS(AST) ((AST)->comp_stmt.numvars)
+#define PKL_AST_COMP_STMT_FRAMELESS_P(AST) ((AST)->comp_stmt.frameless_p)
 
 #define PKL_AST_BUILTIN_NONE 0
-#define PKL_AST_BUILTIN_PRINT 1
-#define PKL_AST_BUILTIN_RAND 2
-#define PKL_AST_BUILTIN_GET_ENDIAN 3
-#define PKL_AST_BUILTIN_SET_ENDIAN 4
-#define PKL_AST_BUILTIN_GET_IOS 5
-#define PKL_AST_BUILTIN_SET_IOS 6
-#define PKL_AST_BUILTIN_OPEN 7
-#define PKL_AST_BUILTIN_CLOSE 8
-#define PKL_AST_BUILTIN_IOSIZE 9
-#define PKL_AST_BUILTIN_GETENV 10
-#define PKL_AST_BUILTIN_FORGET 11
-#define PKL_AST_BUILTIN_GET_TIME 12
-#define PKL_AST_BUILTIN_STRACE 13
 #define PKL_AST_BUILTIN_TERM_GET_COLOR 14
 #define PKL_AST_BUILTIN_TERM_SET_COLOR 15
 #define PKL_AST_BUILTIN_TERM_GET_BGCOLOR 16
 #define PKL_AST_BUILTIN_TERM_SET_BGCOLOR 17
-#define PKL_AST_BUILTIN_TERM_BEGIN_CLASS 19
-#define PKL_AST_BUILTIN_TERM_END_CLASS 20
-#define PKL_AST_BUILTIN_TERM_BEGIN_HYPERLINK 21
-#define PKL_AST_BUILTIN_TERM_END_HYPERLINK 22
-#define PKL_AST_BUILTIN_IOFLAGS 23
-#define PKL_AST_BUILTIN_SLEEP 24
-#define PKL_AST_BUILTIN_IOGETB 25
-#define PKL_AST_BUILTIN_IOSETB 26
-#define PKL_AST_BUILTIN_VM_OBASE 27
-#define PKL_AST_BUILTIN_VM_SET_OBASE 28
-#define PKL_AST_BUILTIN_VM_OPPRINT 29
-#define PKL_AST_BUILTIN_VM_SET_OPPRINT 30
-#define PKL_AST_BUILTIN_VM_OACUTOFF 31
-#define PKL_AST_BUILTIN_VM_SET_OACUTOFF 32
-#define PKL_AST_BUILTIN_VM_ODEPTH 33
-#define PKL_AST_BUILTIN_VM_SET_ODEPTH 34
-#define PKL_AST_BUILTIN_VM_OINDENT 35
-#define PKL_AST_BUILTIN_VM_SET_OINDENT 36
-#define PKL_AST_BUILTIN_VM_OMAPS 37
-#define PKL_AST_BUILTIN_VM_SET_OMAPS 38
-#define PKL_AST_BUILTIN_VM_OMODE 39
-#define PKL_AST_BUILTIN_VM_SET_OMODE 40
-#define PKL_AST_BUILTIN_UNSAFE_STRING_SET 41
 
 struct pkl_ast_comp_stmt
 {
@@ -1587,6 +1673,7 @@ struct pkl_ast_comp_stmt
   union pkl_ast_node *stmts;
   int builtin;
   int numvars;
+  int frameless_p;
 };
 
 pkl_ast_node pkl_ast_make_comp_stmt (pkl_ast ast, pkl_ast_node stmts);
@@ -1607,7 +1694,12 @@ pkl_ast_node pkl_ast_make_null_stmt (pkl_ast ast);
    language.
 
    LVALUE is the l-value of the assignment.
-   EXP is the r-value of the assignment.  */
+   EXP is the r-value of the assignment.
+
+   Note that EXP may be NULL, in which case the assignment statement
+   assumes there is already a r-value on the stack.  This is used to
+   implement the outputs of asm statements.  See
+   pkl_trans1_ps_asm_stmt.  */
 
 #define PKL_AST_ASS_STMT_LVALUE(AST) ((AST)->ass_stmt.lvalue)
 #define PKL_AST_ASS_STMT_EXP(AST) ((AST)->ass_stmt.exp)
@@ -1734,9 +1826,6 @@ pkl_ast_node pkl_ast_make_loop_stmt_iterator (pkl_ast ast,
 
    EXP is the expression to return to the caller.
 
-   NFRAMES is the number of lexical frames to pop before returning
-   from the function.  This is used by the code generator.
-
    NDROPS is the number of stack elements to drop before returning
    from the function.
 
@@ -1747,7 +1836,6 @@ pkl_ast_node pkl_ast_make_loop_stmt_iterator (pkl_ast ast,
    statement. */
 
 #define PKL_AST_RETURN_STMT_EXP(AST) ((AST)->return_stmt.exp)
-#define PKL_AST_RETURN_STMT_NFRAMES(AST) ((AST)->return_stmt.nframes)
 #define PKL_AST_RETURN_STMT_NDROPS(AST) ((AST)->return_stmt.ndrops)
 #define PKL_AST_RETURN_STMT_NPOPES(AST) ((AST)->return_stmt.npopes)
 #define PKL_AST_RETURN_STMT_FUNCTION(AST) ((AST)->return_stmt.function)
@@ -1758,15 +1846,12 @@ struct pkl_ast_return_stmt
 
   union pkl_ast_node *exp;
   union pkl_ast_node *function;
-  int nframes;
   int ndrops;
   int npopes;
 };
 
 pkl_ast_node pkl_ast_make_return_stmt (pkl_ast ast, pkl_ast_node exp);
 
-void pkl_ast_finish_returns (pkl_ast_node function);
-
 /* PKL_AST_EXP_STMT nodes represent "expression statements".
 
    EXP is the expression conforming the statement.  */
@@ -1781,63 +1866,89 @@ struct pkl_ast_exp_stmt
 
 pkl_ast_node pkl_ast_make_exp_stmt (pkl_ast ast, pkl_ast_node exp);
 
-/* PKL_AST_TRY_UNTIL_STMT nodes represent try-until statements.
+/* PKL_AST_TRY_STMT nodes represent try-catch and try-until
+   statements, which are used in order to support exception handlers.
 
-   CODE is a statement.
+   KIND is either PKL_AST_TRY_STMT_KIND_CATCH for try-catch
+   statements, or PKL_AST_TRY_STMT_KIND_UNTIL for try-until
+   statements.
 
-   EXP is an expression that should evaluate to a 32-bit signed
-   integer.  CODE will be executed repeteadly until the given
-   exception type is catched, or some other exception is raised.  */
-
-#define PKL_AST_TRY_UNTIL_STMT_CODE(AST) ((AST)->try_until_stmt.code)
-#define PKL_AST_TRY_UNTIL_STMT_EXP(AST) ((AST)->try_until_stmt.exp)
-
-struct pkl_ast_try_until_stmt
-{
-  struct pkl_ast_common common;
-
-  union pkl_ast_node *code;
-  union pkl_ast_node *exp;
-};
-
-pkl_ast_node pkl_ast_make_try_until_stmt (pkl_ast ast,
-                                          pkl_ast_node code, pkl_ast_node exp);
-
-/* PKL_AST_TRY_CATCH_STMT nodes represent try-catch statements, which
-   are used in order to support exception handlers.
-
-   CODE is a statement that is executed.
+   BODY is the body of the statement, a node of type
+   PKL_AST_TRY_STMT_BODY.
 
    HANDLER is a statement that will be executed in case an exception
-   is raised while executing CODE.
+   is raised while executing the BODY in a try-catch statement.
 
-   TYPE, if specified, is the argument to the catch clause.  The type
-   of the argument must be an Exception.
+   ARG, if specified, is the argument to the catch clause in a
+   try-catch statement.  The type of the argument must be an
+   Exception.
 
-   EXP, if specified, is an expression evaluating to an Exception.
-   Exceptions having any other type won't be catched by the `catch'
-   clause of the statement.
+   EXP is an expression evaluating to an Exception.  In try-catch
+   statements exceptions having any other type won't be caught by the
+   `catch' clause of the statement.  In try-until statements the BODY
+   will be executed repeatedly until this exception gets raised.  In
+   try-catch statements this field is optional.
 
-   Note that TYPE and EXP are mutually exclusive.  */
+   Note that TYPE and EXP are mutually exclusive in try-catch
+   statements.  */
 
-#define PKL_AST_TRY_CATCH_STMT_CODE(AST) ((AST)->try_catch_stmt.code)
-#define PKL_AST_TRY_CATCH_STMT_HANDLER(AST) ((AST)->try_catch_stmt.handler)
-#define PKL_AST_TRY_CATCH_STMT_ARG(AST) ((AST)->try_catch_stmt.arg)
-#define PKL_AST_TRY_CATCH_STMT_EXP(AST) ((AST)->try_catch_stmt.exp)
+#define PKL_AST_TRY_STMT_KIND(AST) ((AST)->try_stmt.kind)
+#define PKL_AST_TRY_STMT_BODY(AST) ((AST)->try_stmt.body)
+#define PKL_AST_TRY_STMT_HANDLER(AST) ((AST)->try_stmt.handler)
+#define PKL_AST_TRY_STMT_ARG(AST) ((AST)->try_stmt.arg)
+#define PKL_AST_TRY_STMT_EXP(AST) ((AST)->try_stmt.exp)
 
-struct pkl_ast_try_catch_stmt
+#define PKL_AST_TRY_STMT_KIND_CATCH 0
+#define PKL_AST_TRY_STMT_KIND_UNTIL 1
+
+struct pkl_ast_try_stmt
 {
   struct pkl_ast_common common;
 
-  union pkl_ast_node *code;
+  int kind;
+  union pkl_ast_node *body;
   union pkl_ast_node *handler;
   union pkl_ast_node *arg;
   union pkl_ast_node *exp;
 };
 
-pkl_ast_node pkl_ast_make_try_catch_stmt (pkl_ast ast,
-                                          pkl_ast_node code, pkl_ast_node handler,
-                                          pkl_ast_node arg, pkl_ast_node exp);
+pkl_ast_node pkl_ast_make_try_stmt (pkl_ast ast,
+                                    int kind,
+                                    pkl_ast_node body, pkl_ast_node handler,
+                                    pkl_ast_node arg, pkl_ast_node exp);
+
+/* PKL_AST_TRY_STMT_BODY nodes represent the body of try-catch and
+   try-until statements.
+
+*/
+
+#define PKL_AST_TRY_STMT_BODY_CODE(AST) ((AST)->try_stmt_body.code)
+
+struct pkl_ast_try_stmt_body
+{
+  struct pkl_ast_common common;
+
+  union pkl_ast_node *code;
+};
+
+pkl_ast_node pkl_ast_make_try_stmt_body (pkl_ast ast, pkl_ast_node code);
+
+/* PKL_AST_TRY_STMT_HANDLER nodes represent the handler part of the
+   try-catch statements.
+
+   CODE is a PKL_AST_COMP_STMT node, or NULL if the node corresponds
+   to a try-until statement.  */
+
+#define PKL_AST_TRY_STMT_HANDLER_CODE(AST) ((AST)->try_stmt_handler.code)
+
+struct pkl_ast_try_stmt_handler
+{
+  struct pkl_ast_common common;
+
+  union pkl_ast_node *code;
+};
+
+pkl_ast_node pkl_ast_make_try_stmt_handler (pkl_ast ast, pkl_ast_node code);
 
 /* PKL_AST_PRINT_STMT nodes represent `print' and `printf' statements.
 
@@ -1862,48 +1973,33 @@ struct pkl_ast_print_stmt
 pkl_ast_node pkl_ast_make_print_stmt (pkl_ast ast,
                                       int printf_p, pkl_ast_node fmt);
 
-/* PKL_AST_CONTINUE_STMT nodes represent `continue' statements.  Each
-   continue statement is associated to a loop node.
+/* PKL_AST_BREAK_CONTINUE_STMT nodes represent `continue' and `break'
+   statements.  Each such statement is associated to a loop node.
 
-   ENTITY is the loop node associated with this continue statement.
+   KIND determines that kind of statement this is.
 
-   NFRAMES is the lexical depth of the continue statement, relative to
-   the enclosing entity.  */
+   ENTITY is the loop node associated with this statement.
 
-#define PKL_AST_CONTINUE_STMT_ENTITY(AST) ((AST)->continue_stmt.entity)
-#define PKL_AST_CONTINUE_STMT_NFRAMES(AST) ((AST)->continue_stmt.nframes)
+   NFRAMES is the lexical depth of the statement, relative to the
+   enclosing entity.  */
 
-struct pkl_ast_continue_stmt
+#define PKL_AST_BREAK_CONTINUE_STMT_KIND(AST) ((AST)->break_continue_stmt.kind)
+#define PKL_AST_BREAK_CONTINUE_STMT_ENTITY(AST) ((AST)->break_continue_stmt.entity)
+#define PKL_AST_BREAK_CONTINUE_STMT_NFRAMES(AST) ((AST)->break_continue_stmt.nframes)
+
+#define PKL_AST_BREAK_CONTINUE_STMT_KIND_BREAK 0
+#define PKL_AST_BREAK_CONTINUE_STMT_KIND_CONTINUE 1
+
+struct pkl_ast_break_continue_stmt
 {
   struct pkl_ast_common common;
   union pkl_ast_node *entity;
+  int kind;
   int nframes;
 };
 
-pkl_ast_node pkl_ast_make_continue_stmt (pkl_ast ast);
-
-/* PKL_AST_BREAK_STMT nodes represent `break' statements.  Each break
-   statement is associated to a loop or switch node.
-
-   ENTITY is the loop or switch node associated with this break
-   statement.
-
-   NFRAMES is the lexical depth of the break statement, relative to
-   the enclosing entity.  */
-
-#define PKL_AST_BREAK_STMT_ENTITY(AST) ((AST)->break_stmt.entity)
-#define PKL_AST_BREAK_STMT_NFRAMES(AST) ((AST)->break_stmt.nframes)
-
-struct pkl_ast_break_stmt
-{
-  struct pkl_ast_common common;
-  union pkl_ast_node *entity;
-  int nframes;
-};
-
-pkl_ast_node pkl_ast_make_break_stmt (pkl_ast ast);
-
-void pkl_ast_finish_breaks (pkl_ast_node entity, pkl_ast_node stmt);
+pkl_ast_node pkl_ast_make_break_continue_stmt (pkl_ast ast,
+                                               int kind);
 
 /* PKL_AST_RAISE_STMT nodes represent `raise' statements, which are
    used in order to raise exceptions at the program level.
@@ -1921,6 +2017,35 @@ struct pkl_ast_raise_stmt
 
 pkl_ast_node pkl_ast_make_raise_stmt (pkl_ast ast, pkl_ast_node exp);
 
+/* PKL_AST_ASM nodes represent `asm' statements, which are used in
+   order to inline PVM assembly in Poke programs.
+
+   TEMPLATE is a string node contaning assembler instructions.
+
+   INPUTS is a chain of expressions.  The result of these expressions
+   will be pushed in the stack in the given order before executing
+   TEMPLATE.
+
+   OUTPUTS is a chain of l-values.  The same number of values on the
+   stack will be assigned to them after executing TEMPLATE, in reverse
+   order.  */
+
+#define PKL_AST_ASM_STMT_TEMPLATE(AST) ((AST)->asm_stmt.template)
+#define PKL_AST_ASM_STMT_INPUTS(AST) ((AST)->asm_stmt.inputs)
+#define PKL_AST_ASM_STMT_OUTPUTS(AST) ((AST)->asm_stmt.outputs)
+
+struct pkl_ast_asm_stmt
+{
+  struct pkl_ast_common common;
+
+  union pkl_ast_node *template;
+  union pkl_ast_node *inputs;
+  union pkl_ast_node *outputs;
+};
+
+pkl_ast_node pkl_ast_make_asm_stmt (pkl_ast ast, pkl_ast_node template,
+                                    pkl_ast_node inputs, pkl_ast_node outputs);
+
 /* Finally, the `pkl_ast_node' type, which represents an AST node of
    any type.  */
 
@@ -1973,12 +2098,14 @@ union pkl_ast_node
   struct pkl_ast_loop_stmt_iterator loop_stmt_iterator;
   struct pkl_ast_return_stmt return_stmt;
   struct pkl_ast_exp_stmt exp_stmt;
-  struct pkl_ast_try_catch_stmt try_catch_stmt;
-  struct pkl_ast_try_until_stmt try_until_stmt;
-  struct pkl_ast_break_stmt break_stmt;
-  struct pkl_ast_continue_stmt continue_stmt;
+  struct pkl_ast_try_stmt try_stmt;
+  struct pkl_ast_try_stmt_body try_stmt_body;
+  struct pkl_ast_try_stmt_body try_stmt_handler;
+  struct pkl_ast_break_continue_stmt break_continue_stmt;
   struct pkl_ast_raise_stmt raise_stmt;
   struct pkl_ast_print_stmt print_stmt;
+  struct pkl_ast_asm_stmt asm_stmt;
+  struct pkl_ast_asm_exp asm_exp;
 };
 
 static inline pkl_ast_node __attribute__ ((always_inline, warn_unused_result))
