@@ -1,6 +1,6 @@
 /* pkl-pass.c - Support for compiler passes.  */
 
-/* Copyright (C) 2019, 2020, 2021, 2022 Jose E. Marchesi */
+/* Copyright (C) 2019, 2020, 2021, 2022, 2023 Jose E. Marchesi */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,6 +18,7 @@
 
 #include <config.h>
 
+#include "pk-utils.h"
 #include "pkl-pass.h"
 
 #define PKL_CALL_PHASES(CLASS,ORDER,DISCR)                              \
@@ -171,7 +172,7 @@ pkl_call_node_handlers (pkl_compiler compiler,
 #undef PKL_DEF_OP
             default:
               /* Unknown operation code.  */
-              assert (0);
+              PK_UNREACHABLE ();
             }
         }
 
@@ -219,7 +220,7 @@ pkl_call_node_handlers (pkl_compiler compiler,
 #undef PKL_DEF_OP
             default:
               /* Unknown operation code.  */
-              assert (0);
+              PK_UNREACHABLE ();
             }
         }
 
@@ -233,7 +234,7 @@ pkl_call_node_handlers (pkl_compiler compiler,
         }
     }
   else
-    assert (0);
+    PK_UNREACHABLE ();
 
  newnode:
  restart:
@@ -334,11 +335,12 @@ pkl_do_pass_1 (pkl_compiler compiler,
   switch (PKL_AST_CODE (node))
     {
     case PKL_AST_EXP:
-      PKL_PASS (PKL_AST_EXP_OPERAND (node, 0));
-      if (PKL_AST_EXP_NUMOPS (node) == 2)
-        PKL_PASS (PKL_AST_EXP_OPERAND (node, 1));
-
-      break;
+      {
+        int i;
+        for (i = 0; i < PKL_AST_EXP_NUMOPS (node); ++i)
+          PKL_PASS (PKL_AST_EXP_OPERAND (node, i));
+        break;
+      }
     case PKL_AST_PROGRAM:
       if (PKL_AST_PROGRAM_ELEMS (node))
         PKL_PASS_CHAIN (PKL_AST_PROGRAM_ELEMS (node));
@@ -455,7 +457,7 @@ pkl_do_pass_1 (pkl_compiler compiler,
             break;
           default:
             /* Unknown type code.  */
-            assert (0);
+            PK_UNREACHABLE ();
           }
         break;
       }
@@ -469,8 +471,10 @@ pkl_do_pass_1 (pkl_compiler compiler,
         PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_CONSTRAINT (node));
       if (PKL_AST_STRUCT_TYPE_FIELD_INITIALIZER (node))
         PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_INITIALIZER (node));
-      if (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND (node))
-        PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND (node));
+      if (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE (node))
+        PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_PRE (node));
+      if (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST (node))
+        PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_OPTCOND_POST (node));
       if (PKL_AST_STRUCT_TYPE_FIELD_LABEL (node))
         PKL_PASS (PKL_AST_STRUCT_TYPE_FIELD_LABEL (node));
       break;
@@ -510,7 +514,8 @@ pkl_do_pass_1 (pkl_compiler compiler,
         PKL_PASS_CHAIN (PKL_AST_COMP_STMT_STMTS (node));
       break;
     case PKL_AST_ASS_STMT:
-      PKL_PASS (PKL_AST_ASS_STMT_EXP (node));
+      if (PKL_AST_ASS_STMT_EXP (node))
+        PKL_PASS (PKL_AST_ASS_STMT_EXP (node));
       PKL_PASS (PKL_AST_ASS_STMT_LVALUE (node));
       break;
     case PKL_AST_IF_STMT:
@@ -534,8 +539,7 @@ pkl_do_pass_1 (pkl_compiler compiler,
       PKL_PASS (PKL_AST_LOOP_STMT_ITERATOR_CONTAINER (node));
       PKL_PASS (PKL_AST_LOOP_STMT_ITERATOR_DECL (node));
       break;
-    case PKL_AST_BREAK_STMT:
-    case PKL_AST_CONTINUE_STMT:
+    case PKL_AST_BREAK_CONTINUE_STMT:
       break;
     case PKL_AST_RETURN_STMT:
       if (PKL_AST_RETURN_STMT_EXP (node))
@@ -544,22 +548,33 @@ pkl_do_pass_1 (pkl_compiler compiler,
     case PKL_AST_EXP_STMT:
       PKL_PASS (PKL_AST_EXP_STMT_EXP (node));
       break;
-    case PKL_AST_TRY_CATCH_STMT:
-      PKL_PASS (PKL_AST_TRY_CATCH_STMT_CODE (node));
-      PKL_PASS (PKL_AST_TRY_CATCH_STMT_HANDLER (node));
-      if (PKL_AST_TRY_CATCH_STMT_ARG (node))
-        PKL_PASS (PKL_AST_TRY_CATCH_STMT_ARG (node));
-      if (PKL_AST_TRY_CATCH_STMT_EXP (node))
-        PKL_PASS (PKL_AST_TRY_CATCH_STMT_EXP (node));
+    case PKL_AST_TRY_STMT_BODY:
+      PKL_PASS (PKL_AST_TRY_STMT_BODY_CODE (node));
       break;
-    case PKL_AST_TRY_UNTIL_STMT:
-      PKL_PASS (PKL_AST_TRY_UNTIL_STMT_CODE (node));
-      PKL_PASS (PKL_AST_TRY_UNTIL_STMT_EXP (node));
+    case PKL_AST_TRY_STMT_HANDLER:
+      PKL_PASS (PKL_AST_TRY_STMT_HANDLER_CODE (node));
+      break;
+    case PKL_AST_TRY_STMT:
+      PKL_PASS (PKL_AST_TRY_STMT_BODY (node));
+      if (PKL_AST_TRY_STMT_HANDLER (node))
+        PKL_PASS (PKL_AST_TRY_STMT_HANDLER (node));
+      if (PKL_AST_TRY_STMT_ARG (node))
+        PKL_PASS (PKL_AST_TRY_STMT_ARG (node));
+      if (PKL_AST_TRY_STMT_EXP (node))
+        PKL_PASS (PKL_AST_TRY_STMT_EXP (node));
       break;
     case PKL_AST_RAISE_STMT:
       if (PKL_AST_RAISE_STMT_EXP (node))
         PKL_PASS (PKL_AST_RAISE_STMT_EXP (node));
       break;
+    case PKL_AST_ASM_STMT:
+      if (PKL_AST_ASM_STMT_TEMPLATE (node))
+        PKL_PASS (PKL_AST_ASM_STMT_TEMPLATE (node));
+      if (PKL_AST_ASM_STMT_INPUTS (node))
+        PKL_PASS_CHAIN (PKL_AST_ASM_STMT_INPUTS (node));
+      if (PKL_AST_ASM_STMT_OUTPUTS (node))
+        PKL_PASS_CHAIN (PKL_AST_ASM_STMT_OUTPUTS (node));
+      break;
     case PKL_AST_FORMAT_ARG:
       if (PKL_AST_FORMAT_ARG_EXP (node))
         PKL_PASS (PKL_AST_FORMAT_ARG_EXP (node));
@@ -580,6 +595,12 @@ pkl_do_pass_1 (pkl_compiler compiler,
     case PKL_AST_LAMBDA:
       PKL_PASS (PKL_AST_LAMBDA_FUNCTION (node));
       break;
+    case PKL_AST_ASM_EXP:
+      PKL_PASS (PKL_AST_ASM_EXP_TYPE (node));
+      PKL_PASS (PKL_AST_ASM_EXP_TEMPLATE (node));
+      if (PKL_AST_ASM_EXP_INPUTS (node))
+        PKL_PASS_CHAIN (PKL_AST_ASM_EXP_INPUTS (node));
+      break;
     case PKL_AST_INCRDECR:
       PKL_PASS (PKL_AST_INCRDECR_EXP (node));
       if (PKL_AST_INCRDECR_ASS_STMT (node))
@@ -596,7 +617,7 @@ pkl_do_pass_1 (pkl_compiler compiler,
       break;
     default:
       /* Unknown node code.  This kills the poke :'( */
-      assert (0);
+      PK_UNREACHABLE ();
     }
 
   /* Call the post-order handlers from registered phases.  */
